/* client.c: .so preload to forward calls */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/select.h>
#include <dlfcn.h>
#include <linux/fcntl.h>
#include <errno.h>

#include "netproto.h"
#include "ioctls.h"
#include "replfuns.h"

fd_set is_forwarded;
int is_init = 0;
struct sockaddr_in server;
int do_work;

int int_function_placeholder(int p1, int p2, ...) { return -1; }
int (*__ioctl_ptr) (int, int request, ...) = &int_function_placeholder;

const char *fwd_files[] = FWD_FILES;

void client_init()
{
  void *handle;
  char *s;
  char servername[100];
  int port;
  
  is_init = 1;
  FD_ZERO(&is_forwarded);
  
  handle = dlopen("/lib/libc.so.6", RTLD_LAZY);
  if (handle != NULL)
    __ioctl_ptr = dlsym(handle, "ioctl");
  dlclose(handle);
  
  /* test for connection */
  do_work = 0;
  port = 0;
  s = getenv(ENV_NAME);
  if (s == NULL) { 
    s = getenv("DISPLAY"); 
    port = DEFAULT_PORT; 
  }
  if (s == NULL) return;
  strncpy(servername, s, 99);
  servername[99] = 0;
  s = strchr(servername, ':');
  if (s != NULL) {
    *(s++) = 0;
    if (*servername == 0) return;
    if (port == 0) port = atoi(s);
  }  
  if (port == 0) port = DEFAULT_PORT;
  lookup_server(servername, port, &server);
  if ((int)server.sin_addr.s_addr == -1) return;
  do_work = 1;
}

int is_eligible(const char *file)
{
  int i;
  
  for (i=0; fwd_files[i] != NULL; i++)
    if (!strcmp(file, fwd_files[i])) return 1;
  return 0;
}

extern int open (const char *file, int oflag, int perm)
{
  int fd, ret;
  
  if (!is_init) client_init();
  fd = -1;
  if (do_work) {
    ret = is_eligible(file);
    if (ret) {
      fd = connect_to_server(&server);
      if (fd == -1) do_work = 0; /* server unreachable */
    }
  }
  if (do_work && ret) {
    send_int(fd, CALL_OPEN);
    send_string(fd, file);
    send_int(fd, oflag);
    if (oflag & O_CREAT) send_int(fd, perm);
    flush_send(fd);
    if (!recv_int(fd, &ret)) ret = -1;
    if (ret < 0) __libc_close(fd); /* not available on remote host */
    else { /* success */
      FD_SET(fd, &is_forwarded);
      return fd; 
    }
  }
  /* revert to local open */
  if (oflag & O_CREAT)
    return __libc_open(file, oflag, perm);
  else
    return __libc_open(file, oflag, 0);
}

extern int close (int fd)
{
  int ret;
  
  if (!is_init) client_init();
  if (fd<0) { errno = EBADF; return -1; }
  if (do_work && FD_ISSET(fd, &is_forwarded)) {
    send_int(fd, CALL_CLOSE);
    flush_send(fd);
    if (!recv_int(fd, &ret)) ret = -1;
    FD_CLR(fd, &is_forwarded);
    __libc_close(fd);
    return ret;
  }
  else return __libc_close(fd);
}

extern int write (int fd, const void *buf, unsigned int count)
{
  int ret;
  
  if (!is_init) client_init();
  if (fd<0) { errno = EBADF; return -1; }
  if (do_work && FD_ISSET(fd, &is_forwarded)) {
    send_int(fd, CALL_WRITE);
    send_int(fd, (int) count);
    send_data(fd, count, (char *)buf);
    flush_send(fd);
    if (!recv_int(fd, &ret)) ret = -1;
    return ret;
  }
  else return __libc_write(fd, buf, count);
}

extern int read (int fd, void *buf, unsigned int count)
{
  int ret;
  
  if (!is_init) client_init();
  if (fd<0) { errno = EBADF; return -1; }
  if (do_work && FD_ISSET(fd, &is_forwarded)) {
    send_int(fd, CALL_READ);
    send_int(fd, (int) count);
    flush_send(fd);
    if (!recv_int(fd, &ret)) ret = -1;
    if (ret > 0)
      if (!recv_data(fd, ret, (char *)buf)) ret = -1;
    return ret;
  }
  else return __libc_read(fd, buf, count);
}

extern off_t lseek (int fd, off_t offset, int whence)
{
  int ret;
  
  if (!is_init) client_init();
  if (fd<0) { errno = EBADF; return -1; }
  if (do_work && FD_ISSET(fd, &is_forwarded)) {
    send_int(fd, CALL_LSEEK);
    send_int(fd, (int) offset);
    send_int(fd, whence);
    flush_send(fd);
    if (!recv_int(fd, &ret)) ret = -1;
    return (off_t)ret;
  }
  else return __libc_lseek(fd, offset, whence);
}

extern int ioctl (int fd, int request, void *arg)
{
  int ret, props, size;
  
  if (!is_init) client_init();
  if (fd<0) { errno = EBADF; return -1; }
  if (do_work && FD_ISSET(fd, &is_forwarded)) {
    if (!prepare_ioctl_info(request, &props, &size)) { /* unsupported */
      errno = EINVAL;
      return -1;
    }
    send_int(fd, CALL_IOCTL);
    send_int(fd, request);
    if (props & PROPS_WRITE) send_data(fd, size, (char *)arg);
    flush_send(fd);
    if (props & PROPS_READ)  recv_data(fd, size, (char *)arg);
    if (!recv_int(fd, &ret)) ret = -1;
    return ret;
  }
  else return __libc_ioctl(fd, request, arg);
}

